#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
# pylint: disable=too-many-public-methods
""" efi variables """

import re
import struct
import logging
import datetime
import collections

from cryptography import x509

from virt.firmware.efi import guids
from virt.firmware.efi import ucs16
from virt.firmware.efi import certs
from virt.firmware.efi import devpath
from virt.firmware.efi import siglist
from virt.firmware.efi import bootentry

##################################################################################################
# constants

# variable attributes
EFI_VARIABLE_NON_VOLATILE                          = 0x00000001
EFI_VARIABLE_BOOTSERVICE_ACCESS                    = 0x00000002
EFI_VARIABLE_RUNTIME_ACCESS                        = 0x00000004
EFI_VARIABLE_HARDWARE_ERROR_RECORD                 = 0x00000008
EFI_VARIABLE_AUTHENTICATED_WRITE_ACCESS            = 0x00000010  # deprecated
EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS = 0x00000020
EFI_VARIABLE_APPEND_WRITE                          = 0x00000040

EFI_VARIABLE_DEFAULT = (EFI_VARIABLE_NON_VOLATILE |
                        EFI_VARIABLE_BOOTSERVICE_ACCESS)

efivar_defaults = {
    'SecureBoot' : {
        'attr' : (EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS),
        'guid' : guids.EfiGlobalVariable,
    },
    'SecureBootEnable' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.EfiSecureBootEnableDisable,
    },
    'CustomMode' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.EfiCustomModeEnable,
    },
    'PK' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS |
                  EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS),
        'guid' : guids.EfiGlobalVariable,
    },
    'KEK' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS |
                  EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS),
        'guid' : guids.EfiGlobalVariable,
    },
    'db' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS |
                  EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS),
        'guid' : guids.EfiImageSecurityDatabase,
    },
    'dbx' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS |
                  EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS),
        'guid' : guids.EfiImageSecurityDatabase,
    },
    'MokList' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'MokListX' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'MokListTrusted' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'SbatLevel' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'SHIM_DEBUG' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'SHIM_VERBOSE' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'FALLBACK_VERBOSE' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS),
        'guid' : guids.Shim,
    },
    'FB_NO_REBOOT' : {
        'attr' : (EFI_VARIABLE_NON_VOLATILE |
                  EFI_VARIABLE_BOOTSERVICE_ACCESS |
                  EFI_VARIABLE_RUNTIME_ACCESS),
        'guid' : guids.Shim,
    },
}

boot_defaults = {
    'attr' : (EFI_VARIABLE_NON_VOLATILE |
              EFI_VARIABLE_BOOTSERVICE_ACCESS |
              EFI_VARIABLE_RUNTIME_ACCESS),
    'guid' : guids.EfiGlobalVariable,
}

sigdb_names = ("PK", "KEK", "db", "dbx", "MokList", "MokListX", "TlsCaCertificate",
               "MokListRT", "MokListXRT")
bool_names  = ('SecureBootEnable', 'CustomMode')
ascii_names = ('Lang', 'PlatformLang', 'SbatLevel')
blist_names = ('BootOrder', 'BootNext')
dpath_names = ('ConIn', 'ConOut', 'ErrOut')

initial_ts  = datetime.datetime(2010, 1, 1, tzinfo = datetime.timezone.utc)

##################################################################################################

# pylint: disable=too-many-arguments,too-many-instance-attributes
class EfiVar:
    """  class for efi variables """

    # pylint: disable=too-many-positional-arguments
    def __init__(self, name,
                 guid = None,
                 attr = None,
                 data = b'',
                 count = 0,
                 pkidx = 0,
                 authdata = None):
        self.name = name
        self.guid = guid
        self.attr = attr
        self.data = data
        self.count = count
        self.pkidx = pkidx
        self.time = None
        self.sigdb = None

        defaults = efivar_defaults.get(str(name))
        if defaults is None and str(name).startswith('Boot'):
            defaults = boot_defaults

        if self.guid is None:
            if defaults:
                self.guid = guids.parse_str(defaults['guid'])
            else:
                self.guid = guids.parse_str(guids.EfiGlobalVariable)
        if self.attr is None:
            if defaults:
                self.attr = defaults['attr']
            else:
                self.attr = EFI_VARIABLE_DEFAULT

        if authdata:
            self.parse_authdata(authdata)

        if str(self.name) in sigdb_names:
            self.sigdb = siglist.EfiSigDB(self.data)

    def parse_time(self, data, offset):
        """ parse struct EFI_TIME """
        (year, month, day, hour, minute, second, ns, tz, dl) = \
            struct.unpack_from("=HBBBBBxLhBx", data, offset)
        if year:
            self.time = datetime.datetime(year, month, day,
                                          hour, minute, second,
                                          int(ns / 1000),
                                          tzinfo = datetime.timezone.utc)
        else:
            self.time = None

    def parse_authdata(self, authdata):
        """ parse struct EFI_VARIABLE_AUTHENTICATION_2 """
        self.parse_time(authdata, 0)
        (length, revision, certtype) = struct.unpack_from("=LHH", authdata, 16)
        guid = guids.parse_bin(authdata, 24)
        if str(guid) != guids.EfiCertPkcs7:
            raise RuntimeError('no pkcs7 signature')
        self.data = authdata [ 16 + length : ]

    def bytes_time(self):
        """ generate struct EFI_TIME """
        if self.time is None:
            return b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0'
        return struct.pack("=HBBBBBxLhBx",
                           self.time.year, self.time.month, self.time.day,
                           self.time.hour, self.time.minute, self.time.second,
                           self.time.microsecond * 1000,
                           0, 0)

    def update_time(self, ts = None):
        if not self.attr & EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS:
            return
        if ts is None:
            ts = datetime.datetime.now(tz = datetime.timezone.utc)
        if self.time is None or self.time < ts:
            self.time = ts

    def sigdb_clear(self):
        if self.sigdb is None:
            raise RuntimeError
        self.sigdb = siglist.EfiSigDB()
        self.data = bytes(self.sigdb)
        self.update_time(initial_ts)

    def sigdb_set(self, sigdb):
        self.sigdb = sigdb
        self.data = bytes(self.sigdb)
        self.update_time(self.sigdb.get_cert_ts())

    def sigdb_add_cert(self, guid, filename):
        if self.sigdb is None:
            raise RuntimeError
        self.sigdb.add_cert(guid, filename)
        self.data = bytes(self.sigdb)
        self.update_time(self.sigdb.get_cert_ts())

    def sigdb_add_hash(self, guid, hashdata):
        if self.sigdb is None:
            raise RuntimeError
        self.sigdb.add_hash(guid, hashdata)
        self.data = bytes(self.sigdb)
        self.update_time()

    def sigdb_add_dummy(self, guid):
        if self.sigdb is None:
            raise RuntimeError
        self.sigdb.add_dummy(guid)
        self.data = bytes(self.sigdb)
        self.update_time(initial_ts)

    def set_bool(self, value):
        if value:
            self.data = b'\x01'
        else:
            self.data = b'\x00'
        self.update_time()

    def set_uint32(self, value):
        self.data = value.to_bytes(4, byteorder = 'little')
        self.update_time()

    def set_boot_entry(self, attr, title, path, optdata = None):
        entry = bootentry.BootEntry(attr = attr, title = title,
                                    devicepath = path, optdata = optdata)
        self.data = bytes(entry)
        self.update_time()

    def set_boot_next(self, index):
        self.data = struct.pack('=H', index)
        self.update_time()

    def set_boot_order(self, order):
        self.data = b''
        for item in order:
            self.data += struct.pack('=H', item)
        self.update_time()

    def append_boot_order(self, index):
        self.data += struct.pack('=H', index)
        self.update_time()

    def set_from_file(self, filename):
        with open(filename, 'rb') as f:
            self.data = f.read()
        self.update_time()

    def fmt_bool(self):
        if self.data[0]:
            return 'bool: ON'
        return 'bool: off'

    def fmt_ascii(self):
        string = self.data.decode().rstrip('\0')
        string = re.sub(r'\n', r'\\n', string)
        return f'ascii: "{string}"'

    def fmt_boot_entry(self):
        entry = bootentry.BootEntry(data = self.data)
        return f'boot entry: {entry}'

    def fmt_boot_list(self):
        bootlist = []
        for pos in range(len(self.data) >> 1):
            nr = struct.unpack_from('=H', self.data, pos * 2)
            bootlist.append(f'{nr[0]:04x}')
            desc= ", ".join(bootlist)
        return f'boot order: {desc}'

    def fmt_dev_path(self):
        obj = devpath.DevicePath(self.data)
        return f'devpath: {obj}'

    # pylint: disable=too-many-return-statements
    def fmt_data(self):
        name = str(self.name)
        if name in bool_names:
            return self.fmt_bool()
        if name in ascii_names:
            return self.fmt_ascii()
        if name in blist_names:
            return self.fmt_boot_list()
        if name in dpath_names:
            return self.fmt_dev_path()
        if name.startswith('Boot0'):
            return self.fmt_boot_entry()

        if len(self.data) in (1, 2, 4, 8):
            name = {
                1 : 'byte',
                2 : 'word',
                4 : 'dword',
                8 : 'qword',
            }
            n = name[len(self.data)]
            d = bytearray(self.data)
            d.reverse()
            return f'{n}: 0x{d.hex()}'

        return None

# pylint: disable=too-many-public-methods
class EfiVarList(collections.UserDict):
    """  class for efi variable list """

    def create(self, name):
        logging.info('create variable %s', name)
        var = EfiVar(ucs16.from_string(name))
        self[name] = var
        return var

    def delete(self, name):
        if self.get(name):
            logging.info('delete variable: %s', name)
            del self[name]
        else:
            logging.warning('variable %s not found', name)

    def set_bool(self, name, value):
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('set variable %s: %s', name, value)
        var.set_bool(value)

    def set_uint32(self, name, value):
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('set variable %s: %s', name, value)
        var.set_uint32(value)


    def set_boot_entry(self, index, title, path, optdata = None):
        name = f'Boot{index:04X}'
        var = self.get(name)
        if not var:
            var = self.create(name)
        t = ucs16.from_string(title)
        logging.info('set variable %s: %s = %s', name, t, path)
        var.set_boot_entry(bootentry.LOAD_OPTION_ACTIVE,
                           t, path, optdata)

    def add_boot_entry(self, title, path, optdata = None):
        for index in range(0xffff):
            name = f'Boot{index:04X}'
            var = self.get(name)
            if not var:
                self.set_boot_entry(index, title, path, optdata)
                return index
        return None

    def set_boot_next(self, index):
        name = 'BootNext'
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('set variable %s: 0x%04X', name, index)
        var.set_boot_next(index)

    def append_boot_order(self, index):
        name = 'BootOrder'
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('append to variable %s: 0x%04X', name, index)
        var.append_boot_order(index)

    def set_from_file(self, name, filename):
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('set variable %s from file %s', name, filename)
        var.set_from_file(filename)


    def add_cert(self, name, owner, filename, replace = False):
        var = self.get(name)
        if not var:
            var = self.create(name)
        if replace:
            logging.info('clear %s sigdb', name)
            var.sigdb_clear()
        logging.info('add %s cert %s', name, filename)
        var.sigdb_add_cert(guids.parse_str(owner), filename)

    def add_hash(self, name, owner, hashdata):
        var = self.get(name)
        if not var:
            var = self.create(name)
        logging.info('add %s hash %s', name, hashdata)
        var.sigdb_add_hash(guids.parse_str(owner),
                           bytes.fromhex(hashdata))

    def add_dummy_dbx(self, owner):
        var = self.get('dbx')
        if var:
            return
        logging.info("add dummy dbx entry")
        var = self.create('dbx')
        var.sigdb_add_dummy(guids.parse_str(owner))


    def enable_secureboot(self):
        self.add_dummy_dbx(guids.OvmfEnrollDefaultKeys)
        self.set_bool('SecureBootEnable', True)
        self.set_bool('CustomMode', False)

    def enroll_platform_with_cert(self, cert,
                                  guid = guids.OvmfEnrollDefaultKeys,
                                  ms_kek = 'all'):
        self.add_cert('PK', guid, cert, True)
        self.add_cert('KEK', guid, cert, True)
        if ms_kek in ('2011', 'all'):
            self.add_cert('KEK', guids.MicrosoftVendor, certs.MS_KEK_2011, False)
        if ms_kek in ('2023', 'all'):
            self.add_cert('KEK', guids.MicrosoftVendor, certs.MS_KEK_2023, False)
        self.add_dummy_dbx(guids.OvmfEnrollDefaultKeys)

    def enroll_platform_redhat(self, ms_kek = 'all'):
        self.enroll_platform_with_cert(certs.REDHAT_PK, ms_kek = ms_kek)

    def enroll_platform_generate(self, name, ms_kek = 'all'):
        pk = certs.pk_generate(cn = name)
        self.enroll_platform_with_cert(pk.name, ms_kek = ms_kek)

    def add_microsoft_keys(self):
        self.add_cert('db', guids.MicrosoftVendor, certs.MS_WIN_2011, False)
        self.add_cert('db', guids.MicrosoftVendor, certs.MS_WIN_2023, False)
        self.add_cert('db', guids.MicrosoftVendor, certs.MS_3RD_2011, False)
        self.add_cert('db', guids.MicrosoftVendor, certs.MS_3RD_2023, False)

    def add_distro_keys(self, distro):
        certlist = certs.DISTRO_CA.get(distro).get('certs')
        if not certlist:
            l = ', '.join(certs.DISTRO_CA.keys())
            raise RuntimeError(f'unknown distro: {distro} (valid: {l})')
        for item in certlist:
            self.add_cert('db', guids.OvmfEnrollDefaultKeys, item, False)


    @staticmethod
    def print_siglist(slist):
        name = guids.name(slist.guid)
        count = len(slist)
        print(f'  siglist type={name} count={count}')
        if slist.x509:
            try:
                scn = slist.x509.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
                print(f'    subject CN={scn.value}')
            except IndexError:
                print('    no subject CN')
            try:
                icn = slist.x509.issuer.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
                print(f'    issuer CN={icn.value}')
            except IndexError:
                print('    no issuer CN')
        elif str(slist.guid) == guids.EfiCertSha256:
            for sig in list(slist):
                print(f"    {sig['data'].hex()}")

    @staticmethod
    def print_hexdump(data):
        hstr = ''
        astr = ''
        pos = 0
        count = 0
        while count < len(data) and count < 256:
            hstr += f'{data[count]:02x} '
            if (data[count] > 0x20 and data[count] < 0x7f):
                astr += f'{data[count]:c}'
            else:
                astr += '.'
            count += 1

            if count % 4 == 0:
                hstr += ' '
                astr += ' '

            if count % 16 == 0 or count == len(data):
                print(f'  {pos:06x}:  {hstr:52s} {astr}')
                hstr = ''
                astr = ''
                pos += 16

        if count < len(data):
            print(f'  {pos:06x}:  [ ... ]')

    def print_normal(self, hexdump = False):
        for (key, var) in sorted(self.items()):
            name = str(var.name)
            gname = guids.name(var.guid)
            size = len(var.data)
            line = f'name={name} guid={gname} size={size}'
            if var.count:
                line += f' count={var.count}'
            if var.pkidx:
                line += f' pkidx={var.pkidx}'
            if var.time:
                line += f' time={var.time}'
            print(line)
            line = var.fmt_data()
            if line:
                print('  ' + line)
            if var.sigdb:
                for slist in var.sigdb:
                    self.print_siglist(slist)
            if hexdump:
                self.print_hexdump(var.data)
            print('')

    def print_compact(self):
        for (key, var) in sorted(self.items()):
            name = str(var.name)
            desc = var.fmt_data()
            if not desc:
                desc = f'blob: {len(var.data)} bytes'
            print(f'{name:20s}: {desc}')



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" linux efivarfs varstore parser """
import os
import struct
import logging

from virt.firmware.efi import guids
from virt.firmware.efi import ucs16
from virt.firmware.efi import efivar

# pylint: disable=too-few-public-methods
class LinuxVarStore:
    """  class for linux efivarfs varstore """

    def __init__(self, path = '/sys/firmware/efi/efivars'):
        self.path = path
        self.scan = {}

    def scanadd(self, name, guid):
        if not self.scan.get(guid):
            self.scan[guid] = {}
        self.scan[guid][name] = True

    def scandel(self, name, guid):
        del self.scan[guid][name]

    def scandir(self):
        if self.scan:
            return
        if not os.path.isdir(self.path):
            return
        with os.scandir(self.path) as it:
            for entry in it:
                if not entry.is_file():
                    continue
                if len(entry.name) < 38:
                    continue
                name = entry.name[ : len(entry.name) - 37 ]
                guid = entry.name[ len(entry.name) - 36 : ]
                self.scanadd(name, guid)

    def get_variable(self, name, guid):
        self.scandir()
        if not self.scan.get(guid):
            return None
        if not self.scan[guid].get(name):
            return None

        filename = os.path.join(self.path, f'{name}-{guid}')
        try:
            with open(filename, "rb") as f:
                attr = int.from_bytes(f.read(4), byteorder='little', signed=False)
                data = f.read()
                var = efivar.EfiVar(ucs16.from_string(name),
                                    guid = guids.parse_str(guid),
                                    attr = attr, data = data)
            return var
        except OSError:
            return None

    def set_variable(self, var):
        filename = os.path.join(self.path, f'{var.name}-{var.guid}')
        blob = struct.pack('=L', var.attr)
        blob += var.data
        logging.info('updating %s', filename)
        with open(filename, "wb") as f:
            f.write(blob)
        self.scanadd(str(var.name), str(var.guid))

    def del_variable(self, name, guid):
        filename = os.path.join(self.path, f'{name}-{guid}')
        logging.info('removing %s', filename)
        os.unlink(filename)
        self.scandel(name, guid)

    def get_varlist(self, volatile = False):
        self.scandir()
        varlist = efivar.EfiVarList()

        logging.info('reading linux varstore from %s', self.path)
        for (guid, names) in self.scan.items():
            for name in names.keys():
                var = self.get_variable(name, guid)
                if var and (var.attr & efivar.EFI_VARIABLE_NON_VOLATILE or volatile):
                    varlist[name] = var
        return varlist



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" functions to decode information from pe binaries """
import struct

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from virt.firmware.efi import guids
from virt.firmware.efi import siglist

def getcstr(data):
    """ get C string (terminated by null byte) """
    idx = 0
    for b in data:
        if b == 0:
            break
        idx += 1
    return data[:idx]

def pe_string(pe, index):
    """ lookup string in string table (right after symbol table) """
    strtab  = pe.FILE_HEADER.PointerToSymbolTable
    strtab += pe.FILE_HEADER.NumberOfSymbols * 18
    strtab += index
    return getcstr(pe.__data__[strtab:])

def pe_section_name(pe, sec):
    """ decode section name """
    if sec.Name.startswith(b'/'):
        idx = getcstr(sec.Name[1:])
        return pe_string(pe, int(idx))
    return getcstr(sec.Name)

def vendor_cert_sigdb(blob):
    # VENDOR_CERT_FILE -> create signature database
    try:
        crt = x509.load_der_x509_certificate(blob, default_backend())
        sl = siglist.EfiSigList(guid = guids.parse_str(guids.EfiCertX509))
        sl.add_sig(guids.parse_str(guids.Shim),
                        crt.public_bytes(serialization.Encoding.DER))
        sigdb = siglist.EfiSigDB()
        sigdb.append(sl)
        return sigdb
    except ValueError:
        pass

    # VENDOR_DB_FILE -> already is a signature datadase
    sigdb = siglist.EfiSigDB(blob)
    return sigdb

def pe_vendor_cert(section):
    db = None
    dbx = None
    vcert = section.get_data()
    (dbs, dbxs, dbo, dbxo) = struct.unpack_from('<IIII', vcert)
    if dbs:
        db = vcert [ dbo : dbo + dbs ]
    if dbxs:
        dbx = vcert [ dbxo : dbxo + dbxs ]
    return (db, dbx)



import os
import json
import argparse
import unittest
import subprocess

from virt.firmware.efi import certs
from virt.firmware.efi import efivar
from virt.firmware.efi import efijson
from virt.firmware.efi import devpath

from virt.firmware.varstore import edk2
from virt.firmware.varstore import linux
from virt.firmware.varstore import aws

from virt.firmware.bootcfg import linuxcfg
from virt.firmware.bootcfg import main as bcfgmain

VARS_EMPTY   = "/usr/share/OVMF/OVMF_VARS.fd"
VARS_SECBOOT = "/usr/share/OVMF/OVMF_VARS.secboot.fd"

TEST_DATA    = os.path.join(os.path.dirname(__file__), "data")
TEST_AWS     = os.path.join(TEST_DATA, 'secboot.aws')
TEST_DBX     = os.path.join(TEST_DATA, 'DBXUpdate-20100307.x64.bin')

cache_detect_container = None

def detect_container():
    global cache_detect_container
    if cache_detect_container is None:
        result = subprocess.run([ 'systemd-detect-virt', '--container', '--quiet' ])
        if result.returncode == 0:
            cache_detect_container = True
        else:
            cache_detect_container = False
    return cache_detect_container

class TestsEdk2(unittest.TestCase):

    @unittest.skipUnless(os.path.exists(VARS_EMPTY), 'no empty vars file')
    def test_probe_edk2_good(self):
        self.assertTrue(edk2.Edk2VarStore.probe(VARS_EMPTY))

    def test_probe_edk2_bad(self):
        self.assertFalse(edk2.Edk2VarStore.probe(TEST_AWS))

    def test_probe_aws_good(self):
        self.assertTrue(aws.AwsVarStore.probe(TEST_AWS))

    @unittest.skipUnless(os.path.exists(VARS_EMPTY), 'no empty vars file')
    def test_probe_aws_bad(self):
        self.assertFalse(aws.AwsVarStore.probe(VARS_EMPTY))

    @unittest.skipUnless(os.path.exists(VARS_EMPTY), 'no empty vars file')
    def test_enroll_microsoft(self):
        store = edk2.Edk2VarStore(VARS_EMPTY)
        varlist = store.get_varlist()
        varlist.enroll_platform_redhat()
        varlist.add_microsoft_keys()
        varlist.enable_secureboot()
        blob = store.bytes_varstore(varlist)

    @unittest.skipUnless(os.path.exists(VARS_EMPTY), 'no empty vars file')
    def test_enroll_distro(self):
        store = edk2.Edk2VarStore(VARS_EMPTY)
        varlist = store.get_varlist()
        varlist.enroll_platform_redhat()
        varlist.add_distro_keys('rhel')
        varlist.add_distro_keys('fedora')
        varlist.add_distro_keys('centos')
        varlist.enable_secureboot()
        blob = store.bytes_varstore(varlist)

    @unittest.skipUnless(os.path.exists(VARS_SECBOOT), 'no secboot vars file')
    def test_json(self):
        store = edk2.Edk2VarStore(VARS_SECBOOT)
        varlist = store.get_varlist()
        j = json.dumps(varlist, cls=efijson.EfiJSONEncoder, indent = 4)
        l = json.loads(j, object_hook = efijson.efi_decode)

    def test_add_hash(self):
        varlist = efivar.EfiVarList()
        varlist.add_hash('db', 'shim', '70183c6c50978ee60f61d8a60580d5e0022114f20f3b99715617054e916770a4')

    def test_set_boot_url(self):
        path = devpath.DevicePath.uri("http://server/path/boot.iso")
        varlist = efivar.EfiVarList()
        varlist.set_boot_entry(0x99, 'netboot', path)
        varlist.set_boot_next(0x99)

    def test_set_boot_filepath(self):
        path = devpath.DevicePath.filepath("\\EFI\\fedora\\shimx64.efi")
        varlist = efivar.EfiVarList()
        varlist.set_boot_entry(0x99, 'shim', path)
        varlist.set_boot_next(0x99)

    @unittest.skipUnless(os.path.exists('/sys/firmware/efi/efivars'), 'no efivars fs')
    def test_parse_linux(self):
        store = linux.LinuxVarStore()
        varlist = store.get_varlist()

    def test_parse_aws(self):
        varlist = aws.AwsVarStore(TEST_AWS)

    def test_generate_aws(self):
        varlist = efivar.EfiVarList()
        varlist.enroll_platform_redhat()
        varlist.add_microsoft_keys()
        varlist.enable_secureboot()
        uefidata = aws.AwsVarStore.base64_varstore(varlist)

    def test_parse_dbxupdate(self):
        with open(TEST_DBX, 'rb') as f:
            blob = f.read()
        dbx = efivar.EfiVar('dbx', authdata = blob)

    def test_generate_pk(self):
        certs.pk_generate()

    @unittest.skipIf(detect_container(), 'in container')
    @unittest.skipUnless(os.path.exists('/sys/firmware/efi/efivars'), 'no efivars fs')
    def test_add_uki(self):
        info = linuxcfg.LinuxOsInfo()
        esp  = info.esp_path()
        if esp is None:
            return
        shim = info.shim_path()
        if shim is None:
            return
        cfg  = linuxcfg.LinuxEfiBootConfig()
        options = argparse.Namespace(shim      = shim,
                                     title     = 'test',
                                     adduki    = f'{esp}/EFI/Linux/test.eki',
                                     cmdline   = None,
                                     bootnext  = True,
                                     bootorder = None,
                                     dryrun    = True)
        bcfgmain.add_uki(cfg, options)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" autodetect varstore format, using the probe method """

from virt.firmware.varstore import aws
from virt.firmware.varstore import edk2
from virt.firmware.varstore import jstore

def open_varstore(filename):
    if edk2.Edk2VarStoreQcow2.probe(filename):
        return edk2.Edk2VarStoreQcow2(filename)

    if edk2.Edk2VarStore.probe(filename):
        return edk2.Edk2VarStore(filename)

    if aws.AwsVarStore.probe(filename):
        return aws.AwsVarStore(filename)

    if jstore.JsonVarStore.probe(filename):
        return jstore.JsonVarStore(filename)

    return None



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" experimental efi boot config tool """
import os
import sys
import logging
import argparse

import pefile

from virt.firmware.efi import ucs16
from virt.firmware.efi import guids
from virt.firmware.efi import devpath
from virt.firmware.efi import bootentry

from virt.firmware.bootcfg import bootcfg
from virt.firmware.bootcfg import linuxcfg

from virt.peutils import pesign


########################################################################
# main

def update_next_or_order(cfg, options, nr):
    if options.bootnext:
        cfg.set_boot_next(nr)
        if not options.dryrun:
            cfg.linux_update_next()

    if options.bootorder is not None:
        cfg.set_boot_order(nr, options.bootorder)
        if not options.dryrun:
            cfg.linux_update_order()


def firmware_loads_efi_binary(cfg, efibinary):
    if not cfg.secureboot:
        return True
    if pesign.cryptography_major < 40:
        # can't use pesign.pe_check_cert() -> play safe
        return False
    db = cfg.varstore.get_variable('db', guids.EfiImageSecurityDatabase)
    pe = pefile.PE(efibinary)
    siglist = pesign.pe_type2_signatures(pe)
    if pesign.pe_check_cert(siglist, db):
        return True
    return False


def create_boot_entry(efiuki, options):
    if options.shim:
        efishim = linuxcfg.LinuxEfiFile(options.shim)
        if efishim.device != efiuki.device:
            logging.error('shim and uki are on different filesystems')
            sys.exit(1)
        if options.cmdline:
            optdata = ucs16.from_string(efiuki.efi_filename() + ' ' + options.cmdline)
        else:
            optdata = ucs16.from_string(efiuki.efi_filename())
        entry = bootentry.BootEntry(title = ucs16.from_string(options.title),
                                    attr = bootentry.LOAD_OPTION_ACTIVE,
                                    devicepath = efishim.dev_path_file(),
                                    optdata = bytes(optdata))
    else:
        if options.cmdline:
            optdata = ucs16.from_string(options.cmdline)
            entry = bootentry.BootEntry(title = ucs16.from_string(options.title),
                                        attr = bootentry.LOAD_OPTION_ACTIVE,
                                        devicepath = efiuki.dev_path_file(),
                                        optdata = bytes(optdata))
        else:
            entry = bootentry.BootEntry(title = ucs16.from_string(options.title),
                                        attr = bootentry.LOAD_OPTION_ACTIVE,
                                        devicepath = efiuki.dev_path_file())
    return entry


def add_uki(cfg, options):
    if not options.shim and not firmware_loads_efi_binary(cfg, options.adduki):
        logging.error('shim binary needed but not found or specified')
        sys.exit(1)
    if not options.title:
        logging.error('entry title not specified')
        sys.exit(1)
    if options.cmdline and cfg.secureboot:
        logging.warning('Overriding built-in UKI cmdline is not possible'
                        ' when Secure Boot is enabled')

    efiuki = linuxcfg.LinuxEfiFile(options.adduki)
    nr = cfg.find_uki_entry(efiuki.efi_filename())
    if nr is None:
        nr = cfg.find_devpath_entry(efiuki.dev_path_file())
    if nr is not None:
        logging.info('Entry exists (Boot%04X)', nr)
    else:
        entry = create_boot_entry(efiuki, options)
        logging.info('Create new entry: %s', str(entry))
        nr = cfg.add_entry(entry)
        logging.info('Added entry (Boot%04X)', nr)
        if not options.dryrun:
            cfg.linux_write_entry(nr)

    update_next_or_order(cfg, options, nr)


def update_uki(cfg, options):
    efiuki = linuxcfg.LinuxEfiFile(options.updateuki)
    nr = cfg.find_uki_entry(efiuki.efi_filename())
    if nr is None:
        nr = cfg.find_devpath_entry(efiuki.dev_path_file())
    if nr is None:
        logging.error('No entry found for %s', options.updateuki)
        sys.exit(1)

    update_next_or_order(cfg, options, nr)


def remove_uki(cfg, options):
    efiuki = linuxcfg.LinuxEfiFile(options.removeuki)
    nr = cfg.find_uki_entry(efiuki.efi_filename())
    if nr is None:
        nr = cfg.find_devpath_entry(efiuki.dev_path_file())
    if nr is None:
        logging.warning('No entry found for %s', options.removeuki)
        return

    logging.info('Removing entry (Boot%04X)', nr)
    cfg.remove_entry(nr)
    if not options.dryrun:
        cfg.linux_remove_entry(nr)
        cfg.linux_update_next()
        cfg.linux_update_order()


def boot_success(cfg, options):
    if cfg.bcurr in cfg.blist:
        logging.info('No update needed, BootCurrent is already in BootOrder.')
        return
    logging.info('Add BootCurrent (Boot%04X) to BootOrder', cfg.bcurr)
    cfg.set_boot_order(cfg.bcurr, 0)
    if not options.dryrun:
        cfg.linux_update_order()


def update_boot_csv(cfg, options):
    if not options.shim:
        logging.error('shim binary not specified')
        sys.exit(1)
    efishim  = linuxcfg.LinuxEfiFile(options.shim)
    shimpath = efishim.dev_path_file()

    shimdir  = os.path.dirname(options.shim)
    shimname = os.path.basename(options.shim)
    csvname  = shimname.upper() \
                     .replace('SHIM', 'BOOT') \
                     .replace('EFI','CSV')

    csvdata = ''
    # Shim fallback creates entries and adds them to BootOrder as it reads
    # BOOT.CSV, this means that the existing BootOrder must be parsed in reversed
    # order.
    for nr in reversed(cfg.blist):
        entry = cfg.bentr[nr]
        if not entry:
            continue
        if not entry.devicepath:
            continue
        if entry.devicepath != shimpath:
            continue
        args = ''
        if entry.optdata:
            args = ucs16.from_ucs16(entry.optdata)
        csvdata += f'{shimname},{entry.title},{args} ,Comment\n'

    logging.info('Updating %s/%s', shimdir, csvname)
    with open(f'{shimdir}/{csvname}', 'wb') as f:
        f.write(b'\xff\xfe')
        f.write(csvdata.encode('utf-16le'))


def add_uri(cfg, options):
    if not options.title:
        logging.error('entry title not specified')
        sys.exit(1)

    devicepath = devpath.DevicePath.uri(options.adduri)
    nr = cfg.find_devpath_entry(devicepath)
    if nr is not None:
        logging.info('Entry exists (Boot%04X)', nr)
    else:
        entry = bootentry.BootEntry(title = ucs16.from_string(options.title),
                                    attr = bootentry.LOAD_OPTION_ACTIVE,
                                    devicepath = devicepath)
        logging.info('Create new entry: %s', str(entry))
        nr = cfg.add_entry(entry)
        logging.info('Added entry (Boot%04X)', nr)
        if not options.dryrun:
            cfg.linux_write_entry(nr)

    update_next_or_order(cfg, options, nr)


def remove_entry(cfg, options):
    nr = int(options.removeentry, base = 16)
    logging.info('Removing entry (Boot%04X)', nr)
    cfg.remove_entry(nr)
    if not options.dryrun:
        cfg.linux_remove_entry(nr)
        cfg.linux_update_next()
        cfg.linux_update_order()


def process_boot_order_arg(pos):
    if pos == 'last':
        return -1
    try:
        return int(pos)
    except argparse.ArgumentTypeError as err:
        raise err
    except (TypeError, ValueError) as err:
        raise argparse.ArgumentTypeError("--boot-order must be a number or 'last'") from err

# pylint: disable=too-many-boolean-expressions,too-many-branches,too-many-statements
def main():
    parser = argparse.ArgumentParser(
        description = 'Show and manage uefi boot entries.')

    parser.add_argument('-l', '--loglevel', dest = 'loglevel', type = str, default = 'info',
                        help = 'set loglevel to LEVEL', metavar = 'LEVEL')
    parser.add_argument('--vars', dest = 'varsfile', type = str,
                        help = 'read edk2 vars from FILE', metavar = 'FILE')
    parser.add_argument('--show', dest = 'show',
                        action = 'store_true', default = False,
                        help = 'print boot configuration')
    parser.add_argument('-v', '--verbose', dest = 'verbose',
                        action = 'store_true', default = False,
                        help = 'print more details')

    group = parser.add_argument_group('update unified kernel image (UKI) boot entries')
    group.add_argument('--add-uki', dest = 'adduki', type = str,
                       help = 'add boot entry for UKI image FILE', metavar = 'FILE')
    group.add_argument('--update-uki', dest = 'updateuki', type = str,
                       help = 'update boot entry for UKI image FILE', metavar = 'FILE')
    group.add_argument('--remove-uki', dest = 'removeuki', type = str,
                       help = 'remove boot entry for UKI image FILE', metavar = 'FILE')
    group.add_argument('--cmdline', dest = 'cmdline', type = str,
                       help = 'override UKIs cmdline when adding boot entry '
                       '(ignored when Secure Boot is enabled)', metavar = 'CMDLINE')
    group.add_argument('--boot-ok', '--boot-successful', dest = 'bootok',
                       action = 'store_true', default = False,
                       help = 'boot is successful, add BootCurrent to BootOrder.')
    group.add_argument('--update-csv', dest = 'updatecsv',
                       action = 'store_true', default = False,
                       help = 'update BOOT.CSV')

    group = parser.add_argument_group('update other boot entries')
    group.add_argument('--add-uri', dest = 'adduri', type = str,
                       help = 'add boot entry to netboot URI', metavar = 'URI')
    group.add_argument('--remove-entry', dest = 'removeentry', type = str,
                       help = 'add remove entry NNNN', metavar = 'NNNN')

    group = parser.add_argument_group('options for boot entry updates')
    group.add_argument('--once', '--boot-next', dest = 'bootnext',
                       action = 'store_true', default = False,
                       help = 'boot added/updated entry once (using BootNext)')
    group.add_argument('--boot-order', dest = 'bootorder', type = process_boot_order_arg,
                       help = 'place added/updated entry at POS in BootOrder (0 is first)',
                       metavar = 'POS')
    group.add_argument('--dry-run', dest = 'dryrun',
                       action = 'store_true', default = False,
                       help = 'do not actually update the configuration')
    group.add_argument('--title', dest = 'title', type = str,
                       help = 'label the entry with TITLE', metavar = 'TITLE')
    group.add_argument('--shim', dest = 'shim', type = str,
                       help = 'use shim binary FILE', metavar = 'FILE')

    group = parser.add_argument_group('print system information')
    group.add_argument('--print-loader', dest = 'printloader',
                       action = 'store_true', default = False,
                       help = 'print boot loader name')

    options = parser.parse_args()

    logging.basicConfig(format = '%(levelname)s: %(message)s',
                        level = getattr(logging, options.loglevel.upper()))

    # sanity checks
    if options.varsfile and (options.adduki or
                             options.updateuki or
                             options.removeuki or
                             options.bootok or
                             options.updatecsv or
                             options.adduri or
                             options.removeentry or
                             options.printloader):
        logging.error('operation not supported for edk2 varstores')
        sys.exit(1)

    # read info
    if options.varsfile:
        cfg = bootcfg.VarStoreEfiBootConfig(options.varsfile)
        loader = None
    else:
        cfg = linuxcfg.LinuxEfiBootConfig()
        loader = cfg.linux_loader_info()

    # find shim if needed
    if not options.shim and (options.adduki or
                             options.updatecsv):
        osinfo = linuxcfg.LinuxOsInfo()
        options.shim = osinfo.shim_path()

    # apply updates
    if options.adduki:
        add_uki(cfg, options)
    elif options.updateuki:
        update_uki(cfg, options)
    elif options.removeuki:
        remove_uki(cfg, options)
    elif options.bootok:
        boot_success(cfg, options)
        if options.updatecsv and options.shim:
            update_boot_csv(cfg, options)
    elif options.updatecsv:
        update_boot_csv(cfg, options)
    elif options.adduri:
        add_uri(cfg, options)
    elif options.removeentry:
        remove_entry(cfg, options)
    elif options.printloader:
        if loader:
            print(str(loader))
        else:
            print('unknown')
    else:
        # default action
        options.show = True

    # print info
    if options.show:
        cfg.print_cfg(options.verbose)

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
# pylint: disable=too-many-instance-attributes,consider-iterating-dictionary
""" core EfiBootConfig class to manage boot configuration """
import re
import struct

from virt.firmware.efi import guids
from virt.firmware.efi import ucs16
from virt.firmware.efi import bootentry

from virt.firmware.varstore import autodetect

class EfiBootConfig:
    """ efi boot configuration """

    def __init__(self):
        self.secureboot  = False
        self.bootorder   = None
        self.bootcurrent = None
        self.bootnext    = None
        self.bcurr = None  # parsed BootCurrent
        self.bnext = None  # parsed BootNext
        self.blist = []    # parsed BootOrder
        self.unused = []   # unused BootNNNN entry list
        self.bentr = {}    # parsed BootNNNN entries
        self.bnext_updated = False
        self.blist_updated = False

    def parse_boot_variables(self):
        if self.bootorder:
            self.blist = []
            for pos in range(len(self.bootorder.data) >> 1):
                nr = struct.unpack_from('=H', self.bootorder.data, pos * 2)
                self.blist.append(nr[0])
                self.bentr[nr[0]] = None
        if self.bootcurrent:
            nr = struct.unpack_from('=H', self.bootcurrent.data)
            self.bcurr = nr[0]
            self.bentr[nr[0]] = None
        if self.bootnext:
            nr = struct.unpack_from('=H', self.bootnext.data)
            self.bnext = nr[0]
            self.bentr[nr[0]] = None

    def add_unused_entries(self, names):
        if not names:
            return
        regex = re.compile('Boot([0-9A-Z]{4})')
        for name in names:
            result = regex.match(name)
            if not result:
                continue
            nr = int(result.group(1), 16)
            if nr in self.bentr:
                continue
            self.unused.append(nr)
            self.bentr[nr] = None

    @staticmethod
    def print_optdata(prefix, optdata):
        if len(optdata) >= 4 and optdata[0] != 0 and optdata[1] == 0:
            print(f'{prefix} opt/ucs16: {ucs16.from_ucs16(optdata, 0)}')
        elif len(optdata) == 16:
            print(f'{prefix} opt/guid: {guids.parse_bin(optdata, 0)}')
        else:
            print(f'{prefix} opt/hex: {optdata.hex()}')

    def print_entry(self, nr, verbose):
        entry = self.bentr[nr]
        cstr = 'C' if nr == self.bcurr else ' '
        nstr = 'N' if nr == self.bnext else ' '
        ostr = 'O' if nr in self.blist else ' '
        if not entry:
            print(f'# {cstr} {nstr} {ostr}  -  {nr:04x}  -  [ missing ]')
            return
        print(f'# {cstr} {nstr} {ostr}  -  {nr:04x}  -  {entry.title}')
        if verbose:
            prefix = '#                    ->'
            print(f'{prefix} path: {entry.devicepath}')
            if entry.optdata:
                self.print_optdata(prefix, entry.optdata)
            print('#')

    def print_cfg(self, verbose = False):
        print('# C - BootCurrent, N - BootNext, O - BootOrder')
        print('# --------------------------------------------')
        if self.bcurr is not None and not self.bcurr in self.blist:
            self.print_entry(self.bcurr, verbose)
        if self.bnext is not None and not self.bnext in self.blist and self.bcurr != self.bnext:
            self.print_entry(self.bnext, verbose)
        for nr in self.blist:
            self.print_entry(nr, verbose)
        for nr in self.unused:
            self.print_entry(nr, verbose)

    def find_uki_entry(self, uki):
        for (nr, entry) in self.bentr.items():
            if not entry or not entry.optdata:
                continue
            try:
                optpath = str(ucs16.from_ucs16(entry.optdata, 0)).split()[0]
            except (UnicodeDecodeError, IndexError):
                continue
            else:
                if optpath == str(uki):
                    return nr
        return None

    def find_title_entry(self, title):
        for (nr, entry) in self.bentr.items():
            if not entry:
                continue
            if str(entry.title) == str(title):
                return nr
        return None

    def find_devpath_entry(self, devicepath):
        blob = bytes(devicepath)
        for (nr, entry) in self.bentr.items():
            if not entry:
                continue
            if blob == bytes(entry.devicepath):
                return nr
        return None

    def find_unused_entry(self):
        nr = 0
        while nr in self.bentr:
            nr += 1
        return nr

    def add_entry(self, entry):
        nr = self.find_unused_entry()
        self.bentr[nr] = entry
        self.unused.append(nr)
        return nr

    def get_entry(self, nr):
        if nr >= len(self.bentr):
            return None
        return self.bentr[nr]

    def remove_entry(self, nr):
        del self.bentr[nr]
        if nr == self.bnext:
            self.bnext = None
            self.bnext_updated = True
        if nr in self.blist:
            self.blist = list(filter(lambda x: x != nr, self.blist))
            self.blist_updated = True
        if nr in self.unused:
            self.unused = list(filter(lambda x: x != nr, self.unused))

    def set_boot_next(self, nr):
        self.bnext = nr
        self.bnext_updated = True
        if nr in self.unused:
            self.unused = list(filter(lambda x: x != nr, self.unused))

    def set_boot_order(self, nr, pos):
        self.blist = list(filter(lambda x: x != nr, self.blist))
        if pos < 0:
            pos = len(self.blist) + pos + 1
        self.blist.insert(pos, nr)
        self.blist_updated = True
        if nr in self.unused:
            self.unused = list(filter(lambda x: x != nr, self.unused))


class VarStoreEfiBootConfig(EfiBootConfig):
    """ read efi boot configuration from varstore  """

    def __init__(self, filename):
        super().__init__()
        self.varstore = None
        self.varlist  = None
        self.varstore_init(filename)

    def varstore_init(self, filename):
        self.varstore = autodetect.open_varstore(filename)
        if self.varstore is None:
            return

        self.varlist = self.varstore.get_varlist()
        self.bootcurrent = None
        self.bootorder = self.varlist.get('BootOrder')
        self.bootnext = self.varlist.get('BootNext')
        self.parse_boot_variables()
        self.add_unused_entries(self.varlist.keys())
        for nr in self.bentr.keys():
            var = self.varlist.get(f'Boot{nr:04X}')
            if var:
                self.bentr[nr] = bootentry.BootEntry(data = var.data)

        sb = self.varlist.get('SecureBootEnable')
        if sb and sb.data[0]:
            self.secureboot = True



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" print efi variables """
import sys
import logging
import argparse

from virt.firmware.varstore import linux


##################################################################################################
# main

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--loglevel', dest = 'loglevel', type = str, default = 'info',
                        help = 'set loglevel to LEVEL', metavar = 'LEVEL')
    parser.add_argument('-v', '--verbose', dest = 'verbose',
                        action = 'store_true', default = False,
                        help = 'print varstore verbosely')
    parser.add_argument('-x', '--hexdump', dest = 'hexdump',
                        action = 'store_true', default = False,
                        help = 'print variable hexdumps')
    parser.add_argument('--volatile', dest = 'volatile',
                        action = 'store_true', default = False,
                        help = 'print volatile variables too')
    options = parser.parse_args()

    logging.basicConfig(format = '%(levelname)s: %(message)s',
                        level = getattr(logging, options.loglevel.upper()))

    varstore = linux.LinuxVarStore()
    varlist = varstore.get_varlist(volatile = options.volatile)
    if options.verbose:
        varlist.print_normal(options.hexdump)
    else:
        varlist.print_compact()

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
"""
boot entry decoder

EFI_LOAD_OPTION
"""

import struct

from virt.firmware.efi import ucs16
from virt.firmware.efi import devpath


LOAD_OPTION_ACTIVE           = 0x00000001
LOAD_OPTION_FORCE_RECONNECT  = 0x00000002
LOAD_OPTION_HIDDEN           = 0x00000008

LOAD_OPTION_CATEGORY         = 0x00001F00
LOAD_OPTION_CATEGORY_BOOT    = 0x00000000
LOAD_OPTION_CATEGORY_APP     = 0x00000100


class BootEntry:
    """ class reprsenting an efi boot entry """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, data = None,
                 attr = None, title = None, devicepath = None, optdata = None):
        self.attr = None
        self.title = None
        self.devicepath = None
        self.optdata = None
        if data:
            self.parse(data)
        if attr:
            self.attr = attr
        if title:
            self.title = title
        if devicepath:
            self.devicepath = devicepath
        if optdata:
            self.optdata = optdata

    def parse(self, data):
        (self.attr, pathsize) = struct.unpack_from('=LH', data)
        self.title = ucs16.from_ucs16(data, 6)
        path = data[ self.title.size() + 6 :
                     self.title.size() + 6 + pathsize ]
        self.devicepath = devpath.DevicePath(path)
        optdata = data[ self.title.size() + 6 + pathsize : ]
        if len(optdata):
            self.optdata = optdata

    def __bytes__(self):
        blob = b''
        pathdata = bytes(self.devicepath)
        blob += struct.pack('=LH', self.attr, len(pathdata))
        blob += bytes(self.title)
        blob += pathdata
        if self.optdata:
            blob += self.optdata
        return blob

    def __str__(self):
        string = f'title="{self.title}" devpath={self.devicepath}'
        if self.optdata:
            string += f' optdata={self.optdata.hex()}'
        return string



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" efi guid database and helper functions """

import uuid

Ffs                          = "8c8ce578-8a3d-4f1c-9935-896185c32dd3"
NvData                       = "fff12b8d-7696-4c8b-a985-2747075b4f50"
AuthVars                     = "aaf32c78-947b-439a-a180-2e144ec37792"
LzmaCompress                 = "ee4e5898-3914-4259-9d6e-dc7bd79403cf"
ResetVector                  = "1ba0062e-c779-4582-8566-336ae8f78f09"

OvmfPeiFv                    = "6938079b-b503-4e3d-9d24-b28337a25806"
OvmfDxeFv                    = "7cb8bdc9-f8eb-4f34-aaea-3ee4af6516a1"

EfiGlobalVariable            = "8be4df61-93ca-11d2-aa0d-00e098032b8c"
EfiImageSecurityDatabase     = "d719b2cb-3d3a-4596-a3bc-dad00e67656f"
EfiSecureBootEnableDisable   = "f0a30bc7-af08-4556-99c4-001009c93a44"
EfiCustomModeEnable          = "c076ec0c-7028-4399-a072-71ee5c448b9f"

EfiCertX509                  = "a5c059a1-94e4-4aa7-87b5-ab155c2bf072"
EfiCertSha256                = "c1c41626-504c-4092-aca9-41f936934328"
EfiCertPkcs7                 = "4aafd29d-68df-49ee-8aa9-347d375665a7"

MicrosoftVendor              = "77fa9abd-0359-4d32-bd60-28f4e78f784b"
OvmfEnrollDefaultKeys        = "a0baa8a3-041d-48a8-bc87-c36d121b5e3d"
Shim                         = "605dab50-e046-4300-abb6-3dd810dd8b23"
LoaderInfo                   = "4a67b082-0a4c-41cf-b6c7-440b29bb8c4f"

OvmfGuidList                 = "96b582de-1fb2-45f7-baea-a366c55a082d"
SevHashTableBlock            = "7255371f-3a3b-4b04-927b-1da6efa8d454"
SevSecretBlock               = "4c2eb361-7d9b-4cc3-8081-127c90d3d294"
SevProcessorReset            = "00f771de-1a7e-4fcb-890e-68c77e2fb44e"
OvmfSevMetadataOffset        = "dc886566-984a-4798-a75e-5585a7bf67cc"
TdxMetadataOffset            = "e47a6535-984a-4798-865e-4685a7bf8ec2"

FwMgrCapsule                 = "6dcbd5ed-e82d-4c44-bda1-7194199ad92a"
SignedCapsule                = "4a3ca68b-7723-48fb-803d-578cc1fec44d"

NotValid                     = "ffffffff-ffff-ffff-ffff-ffffffffffff"

name_table = {
    # firmware volumes
    Ffs                                    : "Ffs",
    NvData                                 : "NvData",
    AuthVars                               : "AuthVars",
    LzmaCompress                           : "LzmaCompress",
    ResetVector                            : "ResetVector",

    "9e21fd93-9c72-4c15-8c4b-e77f1db2d792" : "FvMainCompact",
    "df1ccef6-f301-4a63-9661-fc6030dcc880" : "SecMain",

    "48db5e17-707c-472d-91cd-1613e7ef51b0" : "OvmfMainFv",
    OvmfPeiFv                              : "OvmfPeiFv",
    OvmfDxeFv                              : "OvmfDxeFv",
    "763bed0d-de9f-48f5-81f1-3e90e1b1a015" : "OvmfSecFv",

    # variable types
    EfiGlobalVariable                      : "EfiGlobalVariable",
    EfiImageSecurityDatabase               : "EfiImageSecurityDatabase",
    EfiSecureBootEnableDisable             : "EfiSecureBootEnableDisable",
    EfiCustomModeEnable                    : "EfiCustomModeEnable",

    "eb704011-1402-11d3-8e77-00a0c969723b" : "MtcVendor",
    "4c19049f-4137-4dd3-9c10-8b97a83ffdfa" : "EfiMemoryTypeInformation",
    "4b47d616-a8d6-4552-9d44-ccad2e0f4cf9" : "IScsiConfig",
    "d9bee56e-75dc-49d9-b4d7-b534210f637a" : "EfiCertDb",
    "fd2340d0-3dab-4349-a6c7-3b4f12b48eae" : "EfiTlsCaCertificate",

    # protocols (also used for variables)
    "59324945-ec44-4c0d-b1cd-9db139df070c" : "EfiIScsiInitiatorNameProtocol",
    "9fb9a8a1-2f4a-43a6-889c-d0f7b6c47ad5" : "EfiDhcp6ServiceBindingProtocol",
    "5b446ed1-e30b-4faa-871a-3654eca36080" : "EfiIp4Config2Protocol",
    "937fe521-95ae-4d1a-8929-48bcd90ad31a" : "EfiIp6ConfigProtocol",

    # signature list types
    EfiCertX509                            : "EfiCertX509",
    EfiCertSha256                          : "EfiCertSha256",
    EfiCertPkcs7                           : "EfiCertPkcs7",

    # signature owner
    MicrosoftVendor                        : "MicrosoftVendor",
    OvmfEnrollDefaultKeys                  : "OvmfEnrollDefaultKeys",
    Shim                                   : "Shim",
    LoaderInfo                             : "LoaderInfo",

    # ovmf metadata
    OvmfGuidList                           : "OvmfGuidList",
    OvmfSevMetadataOffset                  : "OvmfSevMetadataOffset",
    TdxMetadataOffset                      : "TdxMetadataOffset",
    SevHashTableBlock                      : "SevHashTableBlock",
    SevSecretBlock                         : "SevSecretBlock",
    SevProcessorReset                      : "SevProcessorReset",

    # capsule
    FwMgrCapsule                           : "FwMgrCapsule",
    SignedCapsule                          : "SignedCapsule",

    # misc
    "00000000-0000-0000-0000-000000000000" : "Zero",
    NotValid                               : "NotValid",
}

def name(guid):
    nstr = name_table.get(str(guid), None)
    if nstr is None:
        return str(guid)
    return f'guid:{nstr}'

def from_name(nstr):
    if nstr.startswith('guid:'):
        nstr.replace('guid:', '', 1)
    for (u, n) in name_table.items():
        if n.upper() == nstr.upper():
            return u
    return None

def parse_bin(data, offset):
    return uuid.UUID(bytes_le = data[offset:offset+16])

def parse_str(nstr):
    try:
        return uuid.UUID(f'urn:uuid:{nstr}')
    except ValueError:
        pass
    ustr = from_name(nstr)
    return uuid.UUID(f'urn:uuid:{ustr}')



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" print and edit ovmf varstore files """
import sys
import logging
import argparse

from virt.firmware.varstore import edk2


##################################################################################################
# main

# pylint: disable=too-many-branches,too-many-statements
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--loglevel', dest = 'loglevel', type = str, default = 'info',
                        help = 'set loglevel to LEVEL', metavar = 'LEVEL')
    parser.add_argument('--template', dest = 'template', type = str,
                        help = 'use varstore template FILE', metavar = 'FILE',
                        default = '/usr/share/edk2/ovmf/OVMF_VARS.fd')
    parser.add_argument('--varstore', dest = 'varstores', type = str, action = 'append',
                        help = 'migrate variable store FILE', metavar = 'FILE', default = [])
    parser.add_argument('--guest', dest = 'guests', type = str, action = 'append',
                        help = 'migrate libvirt guest NAME', metavar = 'NAME', default = [])
    options = parser.parse_args()

    logging.basicConfig(format = '%(levelname)s: %(message)s',
                        level = getattr(logging, options.loglevel.upper()))

    newstore = edk2.Edk2VarStore(options.template)

    for guest in options.guests:
        options.varstores.append(f'/var/lib/libvirt/qemu/nvram/{guest}_VARS.fd')

    for varsfile in options.varstores:
        oldstore = edk2.Edk2VarStore(varsfile)
        varlist = oldstore.get_varlist()
        newstore.write_varstore(varsfile, varlist)

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2024 Gerd Hoffmann
#
""" secure boot databases in *.auth files """
import os
import sys
import struct
import logging
import argparse

from virt.firmware.efi import guids
from virt.firmware.efi import ucs16
from virt.firmware.efi import efivar

from virt.firmware.varstore import autodetect

SECUREBOOT_DATABASE_NAMES = ('PK', 'KEK', 'db', 'dbx')

class AuthFilesVarStore:
    """ secure boot databases in *.auth files """

    def __init__(self, dirname = None):
        self.dirname = dirname
        self.varlist = efivar.EfiVarList()
        if self.dirname:
            self.read_files()

    @staticmethod
    def probe(dirname):
        if not os.path.isdir(dirname):
            return False
        if not os.path.isfile(f'{dirname}/PK.auth'):
            return False
        return True

    def read_files(self):
        logging.info('reading secureboot db varstore from %s', self.dirname)
        for name in SECUREBOOT_DATABASE_NAMES:
            filename = f'{self.dirname}/{name}.auth'
            if not os.path.exists(filename):
                continue
            logging.info('reading %s', filename)
            with open(filename, "rb") as f:
                blob = f.read()
            var = efivar.EfiVar(ucs16.from_string(name),
                                authdata = blob)
            self.varlist[name] = var

    def get_varlist(self):
        return self.varlist

    @staticmethod
    def empty_auth_header():
        """ EFI_VARIABLE_AUTHENTICATION_2 """
        length    = 24
        revision  = 0x0200
        certtype  = 0x0EF1  # WIN_CERT_TYPE_EFI_GUID
        pkcs7guid = guids.parse_str(guids.EfiCertPkcs7)
        blob      = b''
        blob     += struct.pack("=LHH", length, revision, certtype)
        blob     += pkcs7guid.bytes_le
        return blob

    @staticmethod
    def write_varstore(dirname, varlist):
        logging.info('writing secureboot db varstore to %s', dirname)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        if not os.path.isdir(dirname):
            raise RuntimeError(f'{dirname} exists but is not a directory')
        dummy_header = AuthFilesVarStore.empty_auth_header()
        for name in SECUREBOOT_DATABASE_NAMES:
            var = varlist.get(name)
            if var is None:
                continue
            filename = f'{dirname}/{name}.auth'
            logging.info('writing %s', filename)
            with open(filename, "wb") as f:
                f.write(var.bytes_time())
                f.write(dummy_header)
                f.write(var.data)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', dest = 'input', type = str)
    parser.add_argument('--outdir', dest = 'dirname', type = str)
    options = parser.parse_args()

    logging.basicConfig(format = '%(levelname)s: %(message)s',
                        level = logging.DEBUG)
    if options.input:
        logging.info('--- read variables ---')
        vs = autodetect.open_varstore(options.input)
        if vs is None:
            logging.error("unknown input file format")
            sys.exit(1)
        vl = vs.get_varlist()
    else:
        logging.info('--- generate variables ---')
        vl = efivar.EfiVarList()
        vl.enroll_platform_redhat()
        vl.add_microsoft_keys()

    if options.dirname:
        logging.info('--- write variable store ---')
        AuthFilesVarStore.write_varstore(options.dirname, vl)
        logging.info('--- read back variable store ---')
        vs = AuthFilesVarStore(dirname = options.dirname)
        logging.info('--- dumb variables ---')
        vs.get_varlist().print_normal()



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" efi boot menu """
import os
import sys
import argparse
import subprocess

from virt.firmware.bootcfg import linuxcfg

def bootmenu(cfg):
    rows = os.get_terminal_size().lines
    cols = os.get_terminal_size().columns
    cmdline = [ 'dialog', '--menu', 'uefi boot menu',
                str(rows - 10), str(cols - 30), str(rows - 18) ]

    for (nr, entry) in cfg.bentr.items():
        cmdline += [ str(nr), str(entry.title) ]

    result = subprocess.run(cmdline, stderr = subprocess.PIPE, check = False)
    subprocess.run(['clear',], check = True)
    if result.returncode != 0:
        return None
    return int(result.stderr.decode())

def main():
    parser = argparse.ArgumentParser(
        description = 'uefi boot menu')
    parser.add_argument('-r', '--reboot', dest = 'reboot',
                        default = False, action = 'store_true',
                        help = 'reboot after picking an entry')
    options = parser.parse_args()

    cfg = linuxcfg.LinuxEfiBootConfig()
    nr = bootmenu(cfg)
    if nr is None:
        return 0

    cfg.set_boot_next(nr)
    try:
        cfg.linux_update_next()
    except PermissionError:
        print('Can not update BootNext (try run as root)')
        return 1

    if options.reboot:
        message = f'reboot into {cfg.bentr[nr].title}'
        subprocess.run(['/usr/sbin/shutdown', '-r', 'now', message], check = True)

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" pe (efi) binary utilities """
import sys
import gzip
import struct
import argparse

import pefile

from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs7

from virt.firmware.efi import guids
from virt.firmware.misc import cert_not_valid_before, cert_not_valid_after
from virt.firmware.varstore import linux

from virt.peutils import pesign
from virt.peutils import pedecode

def is_ca_cert(cert):
    try:
        bc = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.BASIC_CONSTRAINTS)
    except x509.extensions.ExtensionNotFound:
        bc = False
    if bc:
        return bc.value.ca
    return False

def print_cert(cert, ii, verbose = False):
    print(f'# {ii}   certificate')
    if verbose:
        print(f'# {ii}      subject: {cert.subject.rfc4514_string()}')
        print(f'# {ii}      issuer : {cert.issuer.rfc4514_string()}')
        print(f'# {ii}      valid  : {cert_not_valid_before(cert)} -> {cert_not_valid_after(cert)}')
        print(f'# {ii}      CA     : {is_ca_cert(cert)}')
    else:
        scn = pesign.cert_common_name(cert.subject)
        icn = pesign.cert_common_name(cert.issuer)
        print(f'# {ii}      subject CN: {scn}')
        print(f'# {ii}      issuer  CN: {icn}')

def print_vendor_cert(db, ii, verbose = False):
    sigdb = pedecode.vendor_cert_sigdb(db)
    for sl in sigdb:
        if str(sl.guid) == guids.EfiCertX509:
            print_cert(sl.x509, ii, verbose)
        elif str(sl.guid) == guids.EfiCertSha256:
            print(f'# {ii}   sha256')
            print(f'# {ii}      {len(sl)} entries')
        else:
            print(f'# {ii}   {sl.guid}')

def print_sbat_entries(ii, name, data):
    print(f'# {ii}{name}')
    entries = data.decode().rstrip('\n').split('\n')
    for entry in entries:
        print(f'# {ii}   {entry}')

def sig_type2(data, ii, extract = False, verbose = False, varlist = None):
    certs = pkcs7.load_der_pkcs7_certificates(data)
    for cert in certs:
        print_cert(cert, ii, verbose)
        if varlist:
            for var in ('db', 'dbx', 'MokListRT', 'MokListXRT'):
                if pesign.is_cert_in_sigdb(cert, varlist.get(var)):
                    print(f'# {ii}      certificate found in \'{var}\'')
                elif pesign.is_cert_issuer_in_sigdb(cert, varlist.get(var)):
                    print(f'# {ii}      cert issuer found in \'{var}\'')

        if extract:
            scn = pesign.cert_common_name(cert.subject)
            fn = "".join(x for x in scn if x.isalnum()) + '.pem'
            print(f'# {ii}      >>> {fn}')
            with open(fn, 'wb') as f:
                f.write(cert.public_bytes(serialization.Encoding.PEM))

def pe_section_flags(sec):
    r = '-'
    w = '-'
    x = '-'
    if sec.Characteristics & pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ']:
        r = 'r'
    if sec.Characteristics & pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_WRITE']:
        w = 'w'
    if sec.Characteristics & pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_EXECUTE']:
        x = 'x'
    return r + w + x

# pylint: disable=too-many-arguments,too-many-positional-arguments
def pe_print_sigs(filename, pe, indent, extract, verbose, varlist = None):
    i  = f'{"":{indent}s}'
    ii = f'{"":{indent+3}s}'
    sighdr = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
    if sighdr.VirtualAddress and sighdr.Size:
        print(f'# {i}sigdata:'
              f' addr 0x{sighdr.VirtualAddress:08x} +0x{sighdr.Size:08x}')
        sigs = pe.__data__[ sighdr.VirtualAddress :
                            sighdr.VirtualAddress + sighdr.Size ]
        pos = 0
        index = 0
        while pos + 8 < len(sigs):
            (slen, srev, stype) = struct.unpack_from('<LHH', sigs, pos)
            print(f'# {ii}signature: len 0x{slen:x}, type 0x{stype:x}')
            if extract:
                index += 1
                fn = filename.split('/')[-1] + f'.sig{index}'
                print(f'# {ii}>>> {fn}')
                with open(fn, 'wb') as f:
                    f.write(sigs [ pos : pos + slen ])
            if stype == 2:
                sig_type2(sigs [ pos + 8 : pos + slen ],
                          ii, extract, verbose, varlist)
            pos += slen
            pos = (pos + 7) & ~7 # align

def pe_print_header(pe, indent):
    i = f'{"":{indent}s}'
    isize = pe.OPTIONAL_HEADER.SizeOfImage
    hsize = pe.OPTIONAL_HEADER.SizeOfHeaders
    dll   = pe.OPTIONAL_HEADER.DllCharacteristics
    nx    = "yes" if dll & 0x100 else "no"
    print(f'# {i}header: size=0x{hsize:x} imagesize=0x{isize:x} nx-compat={nx}')

def zboot_binary(pe, indent, verbose, varlist = None):
    i = f'{"":{indent}s}'
    (mz, zimg, zoff, zsize, r1, r2, alg) = struct.unpack_from('<I4sIIII8s', pe.get_data())
    if zimg != b'zimg':
        return

    zalg = pedecode.getcstr(alg).decode()
    print(f'# {i}zboot: 0x{zoff:x} +0x{zsize:x} ({zalg})')

    zdata = pe.__data__[ zoff : zoff + zsize ]
    if zalg == 'gzip':
        data = gzip.decompress(zdata)
    else:
        return

    print(f'# {i}   embedded binary')
    try:
        npe = pefile.PE(data = data)
        pe_print_header(npe, indent + 6)
        for nsec in npe.sections:
            pe_print_section(npe, nsec, indent + 6, verbose, varlist)
        pe_print_sigs(None, npe, indent + 6, False, verbose, varlist)
    except pefile.PEFormatError:
        print(f'# {i}      not a PE binary')

def bzimage_binary(pe, indent):
    i = f'{"":{indent}s}'
    (jump, magic, vmin, vmaj, u1, u2, veroff) = struct.unpack_from('<H4sBBIHH',
                                                                   pe.get_data(), 0x200)
    if magic != b'HdrS':
        return
    (u3, setup_secs) = struct.unpack_from('<BB', pe.get_data(), 0x1f0)
    (pl_off, pl_len) = struct.unpack_from('<II', pe.get_data(), 0x248)

    setup_size = setup_secs * 512 + 512
    pl_off += setup_size
    verstr = pedecode.getcstr(pe.__data__[ veroff + 0x200 : ]).decode()

    (zm1, zm2, zm3, zm4) = struct.unpack_from('<BBBB', pe.__data__, pl_off)
    if zm1 == 0x28 and zm2 == 0xb5 and zm3 == 0x2f and zm4 == 0xfd:
        zalg = 'zstd'
    elif zm1 == 0xfd and zm2 == 0x37 and zm3 == 0x7a and zm4 == 0x58:
        zalg = 'xz'
    elif zm1 == 0x1f and zm2 == 0x8b:
        zalg = 'gzip'
    else:
        zalg = f'unknown ({zm1:x},{zm2:x},{zm3:x},{zm4:x})'

    print(f'# {i}bzImage: bootver={vmaj}.{vmin} setup=0x{setup_size:x} '
          f'payload=0x{pl_off:x}+0x{pl_len:x} compress={zalg}')
    print(f'# {i}   {verstr}')

# pylint: disable=too-many-branches
def pe_print_section(pe, sec, indent, verbose, varlist = None):
    i  = f'{"":{indent}s}'
    ii = f'{"":{indent+3}s}'
    secname = pedecode.pe_section_name(pe, sec)
    print(f'# {i}section:'
          f' file 0x{sec.PointerToRawData:08x} +0x{sec.SizeOfRawData:08x}'
          f'  virt 0x{sec.VirtualAddress:08x} +0x{sec.Misc_VirtualSize:08x}'
          f'  {pe_section_flags(sec)} ({secname.decode()})')
    if secname == b'.vendor_cert':
        (db, dbx) = pedecode.pe_vendor_cert(sec)
        if db:
            print(f'# {ii}db')
            print_vendor_cert(db, ii, verbose)
        if dbx:
            print(f'# {ii}dbx')
            print_vendor_cert(dbx, ii, verbose)
    if secname == b'.sbatlevel':
        levels = sec.get_data()
        (version, poff, loff) = struct.unpack_from('<III', levels)
        print_sbat_entries(ii, 'previous', pedecode.getcstr(levels[poff + 4:]))
        print_sbat_entries(ii, 'latest', pedecode.getcstr(levels[loff + 4:]))
    if secname in (b'.sdmagic', b'.data.ident', b'.cmdline',
                    b'.uname', b'.sbat'):
        lines = sec.get_data().decode().rstrip('\n\0')
        for line in lines.split('\n'):
            print(f'# {ii}{line}')
    if secname == b'.osrel':
        osrel = sec.get_data().decode().rstrip('\n\0')
        entries = osrel.split('\n')
        for entry in entries:
            if entry.startswith('PRETTY_NAME'):
                print(f'# {ii}{entry}')
    if secname == b'.linux':
        print(f'# {ii}embedded binary')
        try:
            npe = pefile.PE(data = sec.get_data())
            pe_print_header(npe, indent + 6)
            for nsec in npe.sections:
                pe_print_section(npe, nsec, indent + 6, verbose, varlist)
            zboot_binary(npe, indent + 6, verbose, varlist)
            pe_print_sigs(None, npe, indent + 6, False, verbose, varlist)
        except pefile.PEFormatError:
            print(f'# {ii}   not a PE binary')

def efi_binary(filename, extract = False, verbose = False, varlist = None):
    print(f'# file: {filename}')
    try:
        pe = pefile.PE(filename)
        pe_print_header(pe, 3)
        for sec in pe.sections:
            pe_print_section(pe, sec, 3, verbose)
        zboot_binary(pe, 3, verbose)
        bzimage_binary(pe, 3)
        pe_print_sigs(filename, pe, 3, extract, verbose, varlist)
    except pefile.PEFormatError:
        print('#    not a PE binary')

def read_sig(filename):
    print(f'# <<< {filename} (signature)')
    with open(filename, 'rb') as f:
        blob = f.read()
    while len(blob) & 7:
        blob += b'\0'
    return blob

def efi_addsig(infile, outfile, sigfiles, replace = False):
    print(f'# <<< {infile} (efi binary)')
    pe = pefile.PE(infile)
    sighdr = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
    addr = sighdr.VirtualAddress
    size = sighdr.Size

    if addr:
        print(f'#    addr: 0x{addr:06x} +0x{size:06x} (existing sigs)')
        copy = addr + size
    else:
        addr = len(pe.__data__)
        copy = addr
        soze = 0
        print(f'#    addr: 0x{addr:06x} (no sigs, appending)')

    if size and replace:
        print('#    drop existing sigs')
        copy = addr
        size = 0

    addsigs = b''
    if sigfiles:
        for sigfile in sigfiles:
            blob = read_sig(sigfile)
            print(f'#    add sig (+0x{len(blob):06x})')
            addsigs += blob
            size += len(blob)

    if outfile:
        print(f'# >>> {outfile} (efi binary)')
        with open(outfile, 'wb') as f:
            print(f'#    fixup addr: 0x{addr:06x} +0x{size:06x} ')
            pe.OPTIONAL_HEADER.DATA_DIRECTORY[4].VirtualAddress = addr
            pe.OPTIONAL_HEADER.DATA_DIRECTORY[4].Size = size
            print(f'#    copy: 0x{copy:06x} bytes')
            f.write(pe.write()[ : copy ])
            if len(addsigs):
                print(f'#    addsigs: 0x{len(addsigs):06x} bytes')
                f.write(addsigs)

def pe_dumpinfo():
    parser = argparse.ArgumentParser()
    parser.add_argument("FILES", nargs='*',
                        help="List of PE files to dump")
    options = parser.parse_args()
    for filename in options.FILES:
        print(f'# file: {filename}')
        pe = pefile.PE(filename)
        print(pe.dump_info())
    return 0

def pe_listsigs():
    parser = argparse.ArgumentParser(
        description = 'Print informations about PE/EFI binaries.')

    parser.add_argument('-x', '--extract', dest = 'extract',
                        action = 'store_true', default = False,
                        help = 'also extract signatures and certificates')
    parser.add_argument('-v', '--verbose', dest = 'verbose',
                        action = 'store_true', default = False,
                        help = 'print more certificate details')
    if pesign.cryptography_major >= 40:
        parser.add_argument('--findcert', '--find-cert', dest = 'findcert',
                            action = 'store_true', default = False,
                            help = 'check EFI databases for certs')
    parser.add_argument("FILES", nargs='*',
                        help="List of PE files to dump")
    options = parser.parse_args()

    varlist = None
    if pesign.cryptography_major >= 40 and options.findcert:
        varlist = linux.LinuxVarStore().get_varlist(volatile = True)

    for filename in options.FILES:
        efi_binary(filename, options.extract, options.verbose, varlist)
    return 0

def pe_addsigs():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', dest = 'infile', type = str,
                        help = 'read efi binary from FILE', metavar = 'FILE')
    parser.add_argument('-o', '--output', dest = 'outfile', type = str,
                        help = 'write efi binary to FILE', metavar = 'FILE')
    parser.add_argument('-s', '--addsig', dest = 'addsigs',
                        type = str, action = 'append',
                        help = 'append  detached signature from FILE',
                        metavar = 'FILE')
    parser.add_argument('--replace', dest = 'replace',
                        action = 'store_true', default = False,
                        help = 'replace existing signatures')
    options = parser.parse_args()

    if not options.infile:
        print('missing input file (try --help)')
        return 1

    efi_addsig(options.infile, options.outfile, options.addsigs, options.replace)
    return 0

if __name__ == '__main__':
    sys.exit(pe_listsigs())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" certificate and signature helper functions """
import struct
import hashlib
import logging

import pkg_resources

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.serialization import pkcs7

from virt.firmware.efi import guids

cryptography_version = pkg_resources.get_distribution('cryptography').version
cryptography_major = int(cryptography_version.split('.')[0])

def cert_common_name(cert):
    try:
        scn = cert.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
        return scn.value
    except IndexError:
        return 'no CN'

def pe_authenticode_hash(pe, method = 'sha256'):
    h = hashlib.new(method)
    blob = pe.__data__

    csum_off = pe.OPTIONAL_HEADER.get_file_offset() + 0x40
    hdr_end = pe.OPTIONAL_HEADER.SizeOfHeaders

    # hash header, excluding checksum and security directory
    h.update(blob [ 0 : csum_off ])
    logging.debug('hash 0x%06x -> 0x%06x - header start -> csum', 0, csum_off)
    if pe.OPTIONAL_HEADER.NumberOfRvaAndSizes < 4:
        sec = None
        h.update(blob [ csum_off + 4 : hdr_end ])
        logging.debug('hash 0x%06x -> 0x%06x - header csum -> end', csum_off + 4, hdr_end)
    else:
        sec = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
        sec_off = sec.get_file_offset()
        h.update(blob [ csum_off + 4 : sec_off ])
        h.update(blob [ sec_off + 8 : hdr_end ])
        logging.debug('hash 0x%06x -> 0x%06x - header csum -> sigs', csum_off + 4, sec_off)
        logging.debug('hash 0x%06x -> 0x%06x - header sigs -> end', sec_off + 8, hdr_end)

    # hash sections
    offset = hdr_end
    for section in sorted(pe.sections, key = lambda s: s.PointerToRawData):
        start = section.PointerToRawData
        end = start + section.SizeOfRawData
        name = section.Name.rstrip(b'\0').decode()
        logging.debug('hash 0x%06x -> 0x%06x - section \'%s\'', start, end, name)
        if start != offset:
            logging.error('unexpected section start 0x%06x (expected 0x%06x, section \'%s\')',
                          start, offset, name)
        h.update(blob [ start : end ])
        offset = end

    # hash remaining data
    if sec and sec.Size:
        end = sec.VirtualAddress
    else:
        end = len(blob)
    if offset < end:
        h.update(blob [ offset : end ])
        logging.debug('hash 0x%06x -> 0x%06x - remaining data', offset, end)

    # hash dword padding
    padding = ((end + 7) & ~7) - end
    if padding:
        for i in range(padding):
            h.update(b'\0')
        logging.debug('hash %d padding byte(s)', padding)

    # log signatures and EOF
    if sec and sec.Size:
        start = sec.VirtualAddress
        end = start + sec.Size
        logging.debug('sigs 0x%06x -> 0x%06x', start, end)
    logging.debug('EOF  0x%06x', len(blob))

    return h.digest()

def pe_type2_signatures(pe):
    siglist = []
    sighdr = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
    if sighdr.VirtualAddress and sighdr.Size:
        sigs = pe.__data__[ sighdr.VirtualAddress :
                            sighdr.VirtualAddress + sighdr.Size ]
        pos = 0
        while pos + 8 < len(sigs):
            (slen, srev, stype) = struct.unpack_from('<LHH', sigs, pos)
            if stype == 2:
                siglist.append(sigs [ pos + 8 : pos + slen ])
            pos += slen
            pos = (pos + 7) & ~7 # align
    return siglist

def is_cert_in_sigdb(cert, variable):
    if variable is None:
        return False
    for item in variable.sigdb:
        if item.x509:
            if item.x509 == cert:
                return True
    return False

def is_cert_issuer_in_sigdb(cert, variable):
    if variable is None:
        return False
    for item in variable.sigdb:
        if item.x509:
            try:
                cert.verify_directly_issued_by(item.x509)
                return True
            except (ValueError, TypeError):
                pass
    return False

#
# This does only check whenever one of the certificates in the pkcs7
# signature is found in the given efi variable, either the certificate
# itself or the issuer of the certificate.
#
# The pkcs7 signature itself is NOT verified.
#
# This requires cryptography_major >= 40.
#
def pe_check_cert(siglist, variable):
    if not variable:
        return None
    for sig in siglist:
        sigcerts = pkcs7.load_der_pkcs7_certificates(sig)
        for sigcert in sigcerts:
            for dbcert in variable.sigdb:
                if dbcert.x509:
                    try:
                        sigcert.verify_directly_issued_by(dbcert.x509)
                        return sigcert
                    except (ValueError, TypeError):
                        pass
                    except InvalidSignature:
                        pass
                    if sigcert == dbcert.x509:
                        return sigcert
    return None

def pe_check_hash(digest, variable):
    if not variable:
        return False
    return variable.sigdb.has_sig(guids.EfiCertSha256, digest)
